@@ -351,6 +351,24 @@ class LookupBuiltins(Visitor):
       return t
 
 
+def MaybeSubstituteParameters(base_type, parameters=None):
+  """Substitutes parameters into base_type if the latter has a template."""
+  # Check if `base_type` is a generic type whose type parameters should be
+  # substituted by `parameters` (a "type macro").
+  template = pytd_utils.GetTypeParameters(base_type)
+  if not template:
+    return None
+  if parameters is None:
+    # TODO(rechen): Respect type parameters' bounds and constraints.
+    mapping = {t: pytd.AnythingType() for t in template}
+  elif len(template) != len(parameters):
+    raise ValueError("%s expected %d parameters, got %s" % (
+        pytd_utils.Print(base_type), len(template), len(parameters)))
+  else:
+    mapping = dict(zip(template, parameters))
+  return base_type.Visit(ReplaceTypeParameters(mapping))
+
+
 class LookupExternalTypes(RemoveTypeParametersFromGenericAny):
   """Look up NamedType pointers using a symbol table."""
 
@@ -371,6 +389,7 @@ class LookupExternalTypes(RemoveTypeParametersFromGenericAny):
     self.name = self_name
     self._in_constant = []
     self._alias_name = None
+    self._in_generic_type = []
     self._star_imports = set()
 
   def _ResolveUsingGetattr(self, module_name, module):
@@ -379,7 +398,7 @@ class LookupExternalTypes(RemoveTypeParametersFromGenericAny):
       g = module.Lookup(module_name + ".__getattr__")
     except KeyError:
       return None
-    # TODO(kramm): Make parser.py actually enforce this:
+    # TODO(b/159059370): Make parser.py actually enforce this:
     assert len(g.signatures) == 1
     return g.signatures[0].return_type
 
@@ -397,6 +416,12 @@ class LookupExternalTypes(RemoveTypeParametersFromGenericAny):
     assert self._alias_name
     self._alias_name = None
 
+  def EnterGenericType(self, _):
+    self._in_generic_type.append(True)
+
+  def LeaveGenericType(self, _):
+    self._in_generic_type.pop()
+
   def _LookupModuleRecursive(self, name):
     module_name, cls_prefix = name, ""
     while module_name not in self._module_map and "." in module_name:
@@ -453,6 +478,10 @@ class LookupExternalTypes(RemoveTypeParametersFromGenericAny):
       item = self._ResolveUsingGetattr(module_name, module)
       if item is None:
         raise KeyError("No %s in module %s" % (name, module_name))
+    if not self._in_generic_type and isinstance(item, pytd.Alias):
+      # If `item` contains type parameters and is not inside a GenericType, then
+      # we replace the parameters with Any.
+      item = MaybeSubstituteParameters(item.type) or item
     return pytd.ToType(item, allow_constants=not self._in_constant)
 
   def VisitClassType(self, t):
@@ -463,6 +492,14 @@ class LookupExternalTypes(RemoveTypeParametersFromGenericAny):
     else:
       return new_type
 
+  def VisitGenericType(self, t):
+    if isinstance(t.base_type, (pytd.GenericType, pytd.UnionType)):
+      try:
+        t = MaybeSubstituteParameters(t.base_type, t.parameters) or t
+      except ValueError as e:
+        raise KeyError(str(e))
+    return t
+
   def _ModulePrefix(self):
     return self.name + "." if self.name else ""
 
@@ -875,7 +912,6 @@ class CreateTypeParametersForSignatures(Visitor):
         self._IsIncomplete(self.function_name)):
       # Leave unknown classes and call traces as-is, they'll never be part of
       # the output.
-      # TODO(kramm): We shouldn't run on call traces in the first place.
       return sig
     counter = _CountUnknowns()
     sig.Visit(counter)
@@ -906,10 +942,6 @@ class CreateTypeParametersForSignatures(Visitor):
       return unit
 
 
-# TODO(kramm): The `~unknown` functionality is becoming more important. Should
-#              we have support for this on the pytd level? (That would mean
-#              changing Class.name to a TYPE). Also, should we just use ~X
-#              instead of ~unknownX?
 class RaiseIfContainsUnknown(Visitor):
   """Find any 'unknown' Class or ClassType (not: pytd.AnythingType!) in a class.
 
@@ -1242,6 +1274,7 @@ class AdjustTypeParameters(Visitor):
     self.function_name = None
     self.constant_name = None
     self.all_typeparams = set()
+    self.generic_level = 0
 
   def _GetTemplateItems(self, param):
     """Get a list of template items from a parameter."""
@@ -1364,6 +1397,30 @@ class AdjustTypeParameters(Visitor):
   def LeaveConstant(self, unused_node):
     self.constant_name = None
 
+  def EnterGenericType(self, unused_node):
+    self.generic_level += 1
+
+  def LeaveGenericType(self, unused_node):
+    self.generic_level -= 1
+
+  def EnterCallableType(self, node):
+    self.EnterGenericType(node)
+
+  def LeaveCallableType(self, node):
+    self.LeaveGenericType(node)
+
+  def EnterTupleType(self, node):
+    self.EnterGenericType(node)
+
+  def LeaveTupleType(self, node):
+    self.LeaveGenericType(node)
+
+  def EnterUnionType(self, node):
+    self.EnterGenericType(node)
+
+  def LeaveUnionType(self, node):
+    self.LeaveGenericType(node)
+
   def _GetFullName(self, name):
     return ".".join(n for n in [self.class_name, name] if n)
 
@@ -1372,10 +1429,13 @@ class AdjustTypeParameters(Visitor):
       return self.class_name
     return self._GetFullName(self.function_name)
 
+  def _IsBoundTypeParam(self, node):
+    in_class = self.class_name and node.name in self.class_typeparams
+    return in_class or self.generic_level
+
   def VisitTypeParameter(self, node):
     """Add scopes to type parameters, track unbound params."""
-    if self.constant_name and (not self.class_name or
-                               node.name not in self.class_typeparams):
+    if self.constant_name and not self._IsBoundTypeParam(node):
       raise ContainerError("Unbound type parameter %s in %s" % (
           node.name, self._GetFullName(self.constant_name)))
     scope = self._GetScope(node.name)
